#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "libcodict.h"

static passwd_t *dict_new_passwd(char *);

static int dict_cmp(char *str1, char *str2)
{
	if(bcmp(str1, str2, strlen(str1)) == 0)
		if(bcmp(str2, str1, strlen(str2)) == 0)
			return 1;

	return 0;
}

dict_t *dict_init(void)
{
	dict_t *dict = NULL;

	if((dict = (dict_t *) malloc(sizeof(dict_t))) == NULL)
		perror("malloc"), exit(-1);

	dict->usr = NULL;
	dict->pwds = NULL;
	dict->next = NULL;

	return dict;
}

void dict_destroy(dict_t *dict)
{
	dict_t *dicttmp = NULL;
	passwd_t *passwd1 = NULL;
	passwd_t *passwd2 = NULL;

	do {
		if(dicttmp != NULL)
			free(dict), dict = dicttmp;

		free(dict->usr);

		for(passwd1 = dict->pwds; passwd1 != NULL; passwd1 = passwd2) {
			passwd2 = passwd1->next;
			free(passwd1->str);
			free(passwd1);
		}
	
		dicttmp = dict->next;
	} while(dicttmp != NULL);
}

dict_t *dict_create(char *str)
{
	dict_t *dict1 = NULL;
	dict_t *dict2 = NULL;
	dict_t *dict3 = NULL;

	passwd_t *passwd1 = NULL;
	passwd_t *passwd2 = NULL;

	FILE *fp = NULL;

	if((fp = fopen(str, "r")) == NULL)
		return NULL;

	while((dict2 = dict_read(fp, NULL)) != NULL) {
/*
		for(passwd1 = dict2->pwds; passwd1 != NULL; passwd1 = passwd1->next)
			for(passwd2 = dict2->pwds->next; passwd2 != NULL; passwd2 = passwd2->next)
				if(!strcmp(passwd1->str, passwd2->str))
					dict_del_pwd(dict2, passwd2);
*/

		if(dict1 == NULL)
			dict1 = dict2, dict3 = dict1;

		dict3->next = dict2;
		dict3 = dict2;
	}

	fclose(fp);

	return dict1;
}

void dict_output(FILE *fp, dict_t *dict)
{
	passwd_t *passwd = NULL;

	for(; dict != NULL; dict = dict->next)
		{
			fprintf(fp, "%s:", dict->usr);

			for(passwd = dict->pwds; passwd != NULL; passwd = passwd->next)
				{
					if(passwd == dict->pwds)
						fprintf(fp, "%s", passwd->str);
					else
						fprintf(fp, ",%s", passwd->str);
				}

			fprintf(fp, "\n");
		}
}

dict_t *dict_read(FILE *fp, dict_t *dicts)
{
	char *ptr1 = NULL;
	char *ptr2 = NULL;
	char buf[1024];
	dict_t *dict = NULL;
	passwd_t *passwd1 = NULL;
	passwd_t *passwd2 = NULL;

	do
		{
			if((fgets(buf, 1024, fp)) == NULL)
				return NULL;
		}
	while((!strncmp(buf, "#", 1)) || ((strchr(buf, ':')) == NULL));

	if((ptr1 = strchr(buf, '\r')) != NULL)
		ptr1[0] = '\0';

	if((ptr1 = strchr(buf, '\n')) != NULL)
		ptr1[0] = '\0';

	if((ptr2 = strchr(buf, ':')) != NULL) {
		ptr2[0] = '\0';
		ptr2++;
	}

	dict = dict_add_usr(NULL, buf);
	dict_add_pwd(dict, buf, NULL);

	while((ptr1 = strchr(ptr2, ',')) != NULL) {
		ptr1[0] = '\0';

		if(strcmp(ptr2, dict->usr))
			dict_add_pwd(dict, ptr2, NULL);

		ptr2 = ptr1 + 1;
	}

	if((ptr2[0] != '\0') && (strcmp(dict->usr, ptr2)))
		dict_add_pwd(dict, ptr2, NULL);

	return dict;
}

int dict_del_pwd(dict_t *dict, passwd_t *passwd)
{
	passwd_t *passwd1 = NULL;
	passwd_t *passwd2 = NULL;

	for(passwd1 = dict->pwds; passwd1 != NULL; passwd1 = passwd1->next) {
		if(passwd1 == passwd) {
			if(passwd2 != NULL)
				passwd2->next = passwd1->next;
			else
				dict->pwds = passwd1->next;

			free(passwd1->str);
			free(passwd1);
	
			return 0;
		}

		passwd2 = passwd1;
	}

	return -1;
}

passwd_t *dict_add_pwd(dict_t *dict, char *str, passwd_t *target)
{
	passwd_t *passwd1 = NULL;
	passwd_t *passwd2 = NULL;

	passwd_t *npasswd = NULL;

	if((npasswd = (passwd_t *) malloc(sizeof(passwd_t))) == NULL)
		perror("malloc"), exit(-1);

	if((npasswd->str = (char *) malloc(strlen(str) + 1)) == NULL)
		perror("malloc"), exit(-1);

	strncpy(npasswd->str, str, strlen(str) + 1);
	npasswd->next = NULL;
		
	if(dict->pwds == NULL) {
		dict->pwds = npasswd;
		return npasswd;
	}

	for(passwd1 = dict->pwds; passwd1 != NULL; passwd1 = passwd1->next)
		{
			if(target != NULL) 
				{
					if(passwd1 == target)
						{
							if(target == dict->pwds)
								{
									if(dict->pwds != NULL)
										npasswd->next = dict->pwds;
	
									dict->pwds = npasswd;
									return npasswd;
								}

							npasswd->next = target->next;
							target->next = npasswd;
	
							return npasswd;
						}
				}
			passwd2 = passwd1;
		}

	passwd2->next = npasswd;

	return npasswd;
}

dict_t *dict_add_usr(dict_t *dict, char *str)
{
	dict_t *ndict = dict_init();
	dict_t *tdict = NULL;

	if((ndict->usr = (char *) malloc(strlen(str) + 1)) == NULL)
		perror("malloc"), exit(-1);

	strncpy(ndict->usr, str, strlen(str) + 1);

	if(dict != NULL)
		{
			for(tdict = dict; tdict->next != NULL; tdict = tdict->next);

			tdict->next = ndict;
		}

	return ndict;
}

int dict_del_usr(dict_t *dict, dict_t *ddict)
{
	dict_t *tdict = NULL;

	for(tdict = dict; tdict != NULL; tdict = tdict->next)
		{
			if(tdict->next == ddict)
				{
					tdict->next = ddict->next;
					ddict->next = NULL;

					dict_destroy(ddict);
					return 0;
				}
		}

	return -1;
}
static passwd_t *dict_new_passwd(char *str)
{
	passwd_t *passwd = NULL;

	if((passwd = (passwd_t *) malloc(sizeof(passwd_t))) == NULL)
		perror("malloc"), exit(-1);
	if((passwd->str = (char *) malloc(strlen(str) + 1)) == NULL)
		perror("malloc"), exit(-1);

	strncpy(passwd->str, str, strlen(str) + 1);
	passwd->next = NULL;

	return passwd;
}

dict_t *dict_clone(dict_t *users)
{
	dict_t *user = NULL;
	dict_t *nuser = NULL;
	dict_t *suser = NULL;
	passwd_t *passwd = NULL;
	passwd_t *npasswd = NULL;

	for(user = users; user != NULL; user = user->next)
		{
			if(nuser == NULL)
				{
					if((nuser = (dict_t *) malloc(sizeof(dict_t))) == NULL)
						perror("malloc"), exit(-1);
				}
			else
				{
					if((nuser->next = (dict_t *) malloc(sizeof(dict_t))) == NULL)
						perror("malloc"), exit(-1);
					else
						nuser = nuser->next;
				}

			if(suser == NULL)
				suser = nuser;

			nuser->usr = NULL;
			nuser->next = NULL;
			nuser->pwds = NULL;

			if(user->usr != NULL)
				{
					if((nuser->usr = (char *) malloc(strlen(user->usr) + 1)) == NULL)
						perror("malloc"), exit(-1);

					strncpy(nuser->usr, user->usr, strlen(user->usr) + 1);
				}

			npasswd = NULL;

			if(user->pwds != NULL)
				for(passwd = user->pwds; passwd != NULL; passwd = passwd->next)
					{
						if(npasswd == NULL)
							{
								if((npasswd = (passwd_t *) malloc(sizeof(passwd_t))) == NULL)
									perror("malloc"), exit(-1);
								else
									nuser->pwds = npasswd;
							}
						else
							{
								if((npasswd->next = (passwd_t *) malloc(sizeof(passwd_t))) == NULL)
									perror("malloc"), exit(-1);
								else
									npasswd = npasswd->next;
							}

						npasswd->str = NULL;
						npasswd->next = NULL;

						if((npasswd->str = (char *) malloc(strlen(passwd->str) + 1)) == NULL)
							perror("malloc"), exit(-1);

						strncpy(npasswd->str, passwd->str, strlen(passwd->str) + 1);
					}
		}

	return suser;
}

dict_t *dict_merge(dict_t *dict, dict_t *pot)
{
	char buf[256];

	int dup = 0;

	dict_t *pot2 = NULL;
	dict_t *pot3 = NULL;
	dict_t *dict2 = NULL;
	dict_t *dict3 = NULL;
	passwd_t *passwd1 = NULL;
	passwd_t *passwd2 = NULL;

	if(dict->usr == NULL)
		return;

	for(dict2 = dict; dict2 != NULL; dict2 = dict2->next)
		{
			pot3 = NULL;
			
			for(pot2 = pot; pot2 != NULL; pot2 = pot2->next)
				{
					if(pot2->usr == NULL)
						break;
					if(!strcmp(pot2->usr, dict2->usr))
						{
							pot3 = pot2;
							break;
						}
				}

			if(pot3 != NULL)
				{
					for(passwd1 = dict2->pwds; passwd1 != NULL; passwd1 = passwd1->next)
						{
							dup = 0;

							for(passwd2 = pot3->pwds; passwd2 != NULL; passwd2 = passwd2->next)
								if(!(strcmp(passwd1->str, passwd2->str)))
									dup = 1;

							if(dup == 0)
								dict_add_pwd(pot3, passwd1->str, NULL);
						}
				}
			else
				{
					dict3 = dict_add_usr(pot, dict2->usr);
					for(passwd1 = dict2->pwds; passwd1 != NULL; passwd1 = passwd1->next)
						if(strcmp(dict3->usr, passwd1->str))
							dict_add_pwd(dict3, passwd1->str, NULL);
				}
		}

	dict_destroy(dict);

	return pot;
}
